@@ -66,7 +66,7 @@ DUK_LOCAL void duk__debug_do_detach1(duk_heap *heap, duk_int_t reason) {
 	/* heap->dbg_detached_cb: keep */
 	/* heap->dbg_udata: keep */
 	/* heap->dbg_processing: keep on purpose to avoid debugger re-entry in detaching state */
-	heap->dbg_paused = 0;
+	DUK_HEAP_CLEAR_DEBUGGER_PAUSED(heap);
 	heap->dbg_state_dirty = 0;
 	heap->dbg_force_restart = 0;
 	heap->dbg_step_type = 0;
@@ -88,6 +88,12 @@ DUK_LOCAL void duk__debug_do_detach1(duk_heap *heap, duk_int_t reason) {
 DUK_LOCAL void duk__debug_do_detach2(duk_heap *heap) {
 	duk_debug_detached_function detached_cb;
 	void *detached_udata;
+	duk_hthread *thr;
+	duk_context *ctx;
+
+	thr = heap->heap_thread;
+	DUK_ASSERT(thr != NULL);
+	ctx = (duk_context *) thr;
 
 	/* Safe to call multiple times. */
 
@@ -102,7 +108,7 @@ DUK_LOCAL void duk__debug_do_detach2(duk_heap *heap) {
 		 * inside the callback.
 		 */
 		DUK_D(DUK_DPRINT("detached during message loop, delayed call to detached_cb"));
-		detached_cb(detached_udata);
+		detached_cb(ctx, detached_udata);
 	}
 
 	heap->dbg_detaching = 0;
@@ -346,10 +352,10 @@ DUK_LOCAL duk_hstring *duk__debug_read_hstring_raw(duk_hthread *thr, duk_uint32_
 		duk_debug_read_bytes(thr, buf, (duk_size_t) len);
 		duk_push_lstring(ctx, (const char *) buf, (duk_size_t) len);
 	} else {
-		p = (duk_uint8_t *) duk_push_fixed_buffer(ctx, (duk_size_t) len);
+		p = (duk_uint8_t *) duk_push_fixed_buffer(ctx, (duk_size_t) len);  /* zero for paranoia */
 		DUK_ASSERT(p != NULL);
 		duk_debug_read_bytes(thr, p, (duk_size_t) len);
-		duk_to_string(ctx, -1);
+		(void) duk_buffer_to_string(ctx, -1);  /* Safety relies on debug client, which is OK. */
 	}
 
 	return duk_require_hstring(ctx, -1);
@@ -379,7 +385,7 @@ DUK_INTERNAL duk_hstring *duk_debug_read_hstring(duk_hthread *thr) {
  fail:
 	DUK_D(DUK_DPRINT("debug connection error: failed to decode int"));
 	DUK__SET_CONN_BROKEN(thr, 1);
-	duk_push_hstring_stridx(thr, DUK_STRIDX_EMPTY_STRING);  /* always push some string */
+	duk_push_hstring_empty(ctx);  /* always push some string */
 	return duk_require_hstring(ctx, -1);
 }
 
@@ -387,7 +393,7 @@ DUK_LOCAL duk_hbuffer *duk__debug_read_hbuffer_raw(duk_hthread *thr, duk_uint32_
 	duk_context *ctx = (duk_context *) thr;
 	duk_uint8_t *p;
 
-	p = (duk_uint8_t *) duk_push_fixed_buffer(ctx, (duk_size_t) len);
+	p = (duk_uint8_t *) duk_push_fixed_buffer(ctx, (duk_size_t) len);  /* zero for paranoia */
 	DUK_ASSERT(p != NULL);
 	duk_debug_read_bytes(thr, p, (duk_size_t) len);
 
@@ -995,7 +1001,7 @@ DUK_INTERNAL void duk_debug_send_status(duk_hthread *thr) {
 	duk_activation *act;
 
 	duk_debug_write_notify(thr, DUK_DBG_CMD_STATUS);
-	duk_debug_write_int(thr, thr->heap->dbg_paused);
+	duk_debug_write_int(thr, (DUK_HEAP_HAS_DEBUGGER_PAUSED(thr->heap) ? 1 : 0));
 
 	DUK_ASSERT_DISABLE(thr->callstack_top >= 0);  /* unsigned */
 	if (thr->callstack_top == 0) {
@@ -1035,15 +1041,15 @@ DUK_INTERNAL void duk_debug_send_throw(duk_hthread *thr, duk_bool_t fatal) {
 	duk_debug_write_int(thr, fatal);
 
 	/* Report thrown value to client coerced to string */
-	duk_dup(ctx, -1);
+	duk_dup_top(ctx);
 	duk__debug_write_hstring_safe_top(thr);
 	duk_pop(ctx);
 
 	if (duk_is_error(ctx, -1)) {
 		/* Error instance, use augmented error data directly */
-		duk_get_prop_stridx(ctx, -1, DUK_STRIDX_FILE_NAME);
+		duk_get_prop_stridx_short(ctx, -1, DUK_STRIDX_FILE_NAME);
 		duk__debug_write_hstring_safe_top(thr);
-		duk_get_prop_stridx(ctx, -2, DUK_STRIDX_LINE_NUMBER);
+		duk_get_prop_stridx_short(ctx, -2, DUK_STRIDX_LINE_NUMBER);
 		duk_debug_write_uint(thr, duk_get_uint(ctx, -1));
 	} else {
 		/* For anything other than an Error instance, we calculate the error
@@ -1149,6 +1155,16 @@ DUK_LOCAL void duk__debug_skip_to_eom(duk_hthread *thr) {
 	}
 }
 
+DUK_LOCAL duk_int32_t duk__debug_read_validate_csindex(duk_hthread *thr) {
+	duk_int32_t level;
+	level = duk_debug_read_int(thr);
+	if (level >= 0 || -level > (duk_int32_t) thr->callstack_top) {
+		duk_debug_write_error_eom(thr, DUK_DBG_ERR_NOTFOUND, "invalid callstack index");
+		return 0;  /* zero indicates failure */
+	}
+	return level;
+}
+
 /*
  *  Simple commands
  */
@@ -1216,7 +1232,7 @@ DUK_LOCAL void duk__debug_handle_step(duk_hthread *thr, duk_heap *heap, duk_int3
 
 	line = duk_debug_curr_line(thr);
 	if (line > 0) {
-		heap->dbg_paused = 0;
+		DUK_HEAP_CLEAR_DEBUGGER_PAUSED(heap);
 		heap->dbg_step_type = step_type;
 		heap->dbg_step_thread = thr;
 		heap->dbg_step_csindex = thr->callstack_top - 1;
@@ -1285,18 +1301,12 @@ DUK_LOCAL void duk__debug_handle_get_var(duk_hthread *thr, duk_heap *heap) {
 	DUK_UNREF(heap);
 	DUK_D(DUK_DPRINT("debug command GetVar"));
 
+	level = duk__debug_read_validate_csindex(thr);
+	if (level == 0) {
+		return;
+	}
 	str = duk_debug_read_hstring(thr);  /* push to stack */
 	DUK_ASSERT(str != NULL);
-	if (duk_debug_peek_byte(thr) != DUK_DBG_IB_EOM) {
-		level = duk_debug_read_int(thr);  /* optional callstack level */
-		if (level >= 0 || -level > (duk_int32_t) thr->callstack_top) {
-			DUK_D(DUK_DPRINT("invalid callstack level for GetVar"));
-			duk_debug_write_error_eom(thr, DUK_DBG_ERR_NOTFOUND, "invalid callstack level");
-			return;
-		}
-	} else {
-		level = -1;
-	}
 
 	if (thr->callstack_top > 0) {
 		rc = duk_js_getvar_activation(thr,
@@ -1332,6 +1342,10 @@ DUK_LOCAL void duk__debug_handle_put_var(duk_hthread *thr, duk_heap *heap) {
 	DUK_UNREF(heap);
 	DUK_D(DUK_DPRINT("debug command PutVar"));
 
+	level = duk__debug_read_validate_csindex(thr);
+	if (level == 0) {
+		return;
+	}
 	str = duk_debug_read_hstring(thr);  /* push to stack */
 	DUK_ASSERT(str != NULL);
 	tv = duk_debug_read_tval(thr);
@@ -1339,16 +1353,6 @@ DUK_LOCAL void duk__debug_handle_put_var(duk_hthread *thr, duk_heap *heap) {
 		/* detached */
 		return;
 	}
-	if (duk_debug_peek_byte(thr) != DUK_DBG_IB_EOM) {
-		level = duk_debug_read_int(thr);  /* optional callstack level */
-		if (level >= 0 || -level > (duk_int32_t) thr->callstack_top) {
-			DUK_D(DUK_DPRINT("invalid callstack level for PutVar"));
-			duk_debug_write_error_eom(thr, DUK_DBG_ERR_NOTFOUND, "invalid callstack level");
-			return;
-		}
-	} else {
-		level = -1;
-	}
 
 	if (thr->callstack_top > 0) {
 		duk_js_putvar_activation(thr,
@@ -1399,9 +1403,9 @@ DUK_LOCAL void duk__debug_handle_get_call_stack(duk_hthread *thr, duk_heap *heap
 			 * value stack operations.
 			 */
 			duk_push_tval(ctx, &curr_act->tv_func);
-			duk_get_prop_stridx(ctx, -1, DUK_STRIDX_FILE_NAME);
+			duk_get_prop_stridx_short(ctx, -1, DUK_STRIDX_FILE_NAME);
 			duk__debug_write_hstring_safe_top(thr);
-			duk_get_prop_stridx(ctx, -2, DUK_STRIDX_NAME);
+			duk_get_prop_stridx_short(ctx, -2, DUK_STRIDX_NAME);
 			duk__debug_write_hstring_safe_top(thr);
 			pc = duk_hthread_get_act_curr_pc(thr, curr_act);
 			if (i != curr_thr->callstack_top - 1 && pc > 0) {
@@ -1428,21 +1432,11 @@ DUK_LOCAL void duk__debug_handle_get_locals(duk_hthread *thr, duk_heap *heap) {
 
 	DUK_UNREF(heap);
 
-	if (duk_debug_peek_byte(thr) != DUK_DBG_IB_EOM) {
-		level = duk_debug_read_int(thr);  /* optional callstack level */
-		if (level >= 0 || -level > (duk_int32_t) thr->callstack_top) {
-			DUK_D(DUK_DPRINT("invalid callstack level for GetLocals"));
-			duk_debug_write_error_eom(thr, DUK_DBG_ERR_NOTFOUND, "invalid callstack level");
-			return;
-		}
-		duk_debug_write_reply(thr);
-	} else {
-		duk_debug_write_reply(thr);
-		if (thr->callstack_top == 0) {
-			goto callstack_empty;
-		}
-		level = -1;
+	level = duk__debug_read_validate_csindex(thr);
+	if (level == 0) {
+		return;
 	}
+	duk_debug_write_reply(thr);
 
 	curr_act = thr->callstack + thr->callstack_top + level;
 
@@ -1453,12 +1447,11 @@ DUK_LOCAL void duk__debug_handle_get_locals(duk_hthread *thr, duk_heap *heap) {
 	 */
 
 	duk_push_tval(ctx, &curr_act->tv_func);
-	duk_get_prop_stridx(ctx, -1, DUK_STRIDX_INT_VARMAP);
+	duk_get_prop_stridx_short(ctx, -1, DUK_STRIDX_INT_VARMAP);
 	if (duk_is_object(ctx, -1)) {
 		duk_enum(ctx, -1, 0 /*enum_flags*/);
 		while (duk_next(ctx, -1 /*enum_index*/, 0 /*get_value*/)) {
-			varname = duk_get_hstring(ctx, -1);
-			DUK_ASSERT(varname != NULL);
+			varname = duk_known_hstring(ctx, -1);
 
 			duk_js_getvar_activation(thr, curr_act, varname, 0 /*throw_flag*/);
 			/* [ ... func varmap enum key value this ] */
@@ -1470,7 +1463,6 @@ DUK_LOCAL void duk__debug_handle_get_locals(duk_hthread *thr, duk_heap *heap) {
 		DUK_D(DUK_DPRINT("varmap is not an object in GetLocals, ignore"));
 	}
 
- callstack_empty:
 	duk_debug_write_eom(thr);
 }
 
@@ -1479,6 +1471,8 @@ DUK_LOCAL void duk__debug_handle_eval(duk_hthread *thr, duk_heap *heap) {
 	duk_small_uint_t call_flags;
 	duk_int_t call_ret;
 	duk_small_int_t eval_err;
+	duk_int_t num_eval_args;
+	duk_bool_t direct_eval;
 	duk_int32_t level;
 
 	DUK_UNREF(heap);
@@ -1489,39 +1483,47 @@ DUK_LOCAL void duk__debug_handle_eval(duk_hthread *thr, duk_heap *heap) {
 	 * activation.  For now, use global object eval() function, with the eval
 	 * considered a 'direct call to eval'.
 	 *
-	 * Callstack level for debug commands only affects scope -- the callstack
+	 * Callstack index for debug commands only affects scope -- the callstack
 	 * as seen by, e.g. Duktape.act() will be the same regardless.
 	 */
 
-	/* nargs == 2 so we can pass a callstack level to eval(). */
+	/* nargs == 2 so we can pass a callstack index to eval(). */
 	duk_push_c_function(ctx, duk_bi_global_object_eval, 2 /*nargs*/);
 	duk_push_undefined(ctx);  /* 'this' binding shouldn't matter here */
 
-	(void) duk_debug_read_hstring(thr);
-	if (duk_debug_peek_byte(thr) != DUK_DBG_IB_EOM) {
-		level = duk_debug_read_int(thr);  /* optional callstack level */
-		if (level >= 0 || -level > (duk_int32_t) thr->callstack_top) {
-			DUK_D(DUK_DPRINT("invalid callstack level for Eval"));
-			duk_debug_write_error_eom(thr, DUK_DBG_ERR_NOTFOUND, "invalid callstack level");
+	/* Read callstack index, if non-null. */
+	if (duk_debug_peek_byte(thr) == DUK_DBG_IB_NULL) {
+		direct_eval = 0;
+		(void) duk_debug_read_byte(thr);
+	} else {
+		direct_eval = 1;
+		level = duk__debug_read_validate_csindex(thr);
+		if (level == 0) {
 			return;
 		}
 	}
-	else {
-		level = -1;
+
+	DUK_ASSERT(!direct_eval ||
+	           (level < 0 && -level <= (duk_int32_t) thr->callstack_top));
+
+	(void) duk_debug_read_hstring(thr);
+	if (direct_eval) {
+		num_eval_args = 2;
+		duk_push_int(ctx, level - 1);  /* compensate for eval() call */
+	} else {
+		num_eval_args = 1;
 	}
-	DUK_ASSERT(level < 0 && -level <= (duk_int32_t) thr->callstack_top);
-	duk_push_int(ctx, level - 1);  /* compensate for eval() call */
 
 	/* [ ... eval "eval" eval_input level ] */
 
 	call_flags = 0;
-	if (thr->callstack_top >= (duk_size_t) -level) {
+	if (direct_eval && thr->callstack_top >= (duk_size_t) -level) {
 		duk_activation *act;
 		duk_hobject *fun;
 
 		act = thr->callstack + thr->callstack_top + level;
 		fun = DUK_ACT_GET_FUNC(act);
-		if (fun != NULL && DUK_HOBJECT_IS_COMPILEDFUNCTION(fun)) {
+		if (fun != NULL && DUK_HOBJECT_IS_COMPFUNC(fun)) {
 			/* Direct eval requires that there's a current
 			 * activation and it is an Ecmascript function.
 			 * When Eval is executed from e.g. cooperate API
@@ -1531,7 +1533,7 @@ DUK_LOCAL void duk__debug_handle_eval(duk_hthread *thr, duk_heap *heap) {
 		}
 	}
 
-	call_ret = duk_handle_call_protected(thr, 2 /*num_stack_args*/, call_flags);
+	call_ret = duk_handle_call_protected(thr, num_eval_args, call_flags);
 
 	if (call_ret == DUK_EXEC_SUCCESS) {
 		eval_err = 0;
@@ -1786,7 +1788,7 @@ DUK_LOCAL void duk__debug_dump_strtab_probe(duk_hthread *thr, duk_heap *heap) {
 
 	for (i = 0; i < heap->st_size; i++) {
 #if defined(DUK_USE_HEAPPTR16)
-		h = DUK_USE_HEAPPTR_DEC16(heap->strtable16[i]);
+		h = DUK_USE_HEAPPTR_DEC16(heap->heap_udata, heap->strtable16[i]);
 #else
 		h = heap->strtable[i];
 #endif
@@ -1816,7 +1818,7 @@ DUK_LOCAL void duk__debug_handle_dump_heap(duk_hthread *thr, duk_heap *heap) {
 
 DUK_LOCAL void duk__debug_handle_get_bytecode(duk_hthread *thr, duk_heap *heap) {
 	duk_activation *act;
-	duk_hcompiledfunction *fun = NULL;
+	duk_hcompfunc *fun = NULL;
 	duk_size_t i, n;
 	duk_tval *tv;
 	duk_hobject **fn;
@@ -1836,7 +1838,7 @@ DUK_LOCAL void duk__debug_handle_get_bytecode(duk_hthread *thr, duk_heap *heap)
 		}
 		if (DUK_TVAL_IS_OBJECT(tv)) {
 			/* tentative, checked later */
-			fun = (duk_hcompiledfunction *) DUK_TVAL_GET_OBJECT(tv);
+			fun = (duk_hcompfunc *) DUK_TVAL_GET_OBJECT(tv);
 			DUK_ASSERT(fun != NULL);
 		} else if (DUK_TVAL_IS_NUMBER(tv)) {
 			level = (duk_int32_t) DUK_TVAL_GET_NUMBER(tv);
@@ -1848,37 +1850,37 @@ DUK_LOCAL void duk__debug_handle_get_bytecode(duk_hthread *thr, duk_heap *heap)
 
 	if (fun == NULL) {
 		if (level >= 0 || -level > (duk_int32_t) thr->callstack_top) {
-			DUK_D(DUK_DPRINT("invalid callstack level for GetBytecode"));
-			goto fail_level;
+			DUK_D(DUK_DPRINT("invalid callstack index for GetBytecode"));
+			goto fail_index;
 		}
 		act = thr->callstack + thr->callstack_top + level;
-		fun = (duk_hcompiledfunction *) DUK_ACT_GET_FUNC(act);
+		fun = (duk_hcompfunc *) DUK_ACT_GET_FUNC(act);
 	}
 
-	if (fun == NULL || !DUK_HOBJECT_IS_COMPILEDFUNCTION((duk_hobject *) fun)) {
+	if (fun == NULL || !DUK_HOBJECT_IS_COMPFUNC((duk_hobject *) fun)) {
 		DUK_D(DUK_DPRINT("invalid argument to GetBytecode: %!O", fun));
 		goto fail_args;
 	}
-	DUK_ASSERT(fun != NULL && DUK_HOBJECT_IS_COMPILEDFUNCTION((duk_hobject *) fun));
+	DUK_ASSERT(fun != NULL && DUK_HOBJECT_IS_COMPFUNC((duk_hobject *) fun));
 
 	duk_debug_write_reply(thr);
-	n = DUK_HCOMPILEDFUNCTION_GET_CONSTS_COUNT(heap, fun);
+	n = DUK_HCOMPFUNC_GET_CONSTS_COUNT(heap, fun);
 	duk_debug_write_int(thr, (duk_int32_t) n);
-	tv = DUK_HCOMPILEDFUNCTION_GET_CONSTS_BASE(heap, fun);
+	tv = DUK_HCOMPFUNC_GET_CONSTS_BASE(heap, fun);
 	for (i = 0; i < n; i++) {
 		duk_debug_write_tval(thr, tv);
 		tv++;
 	}
-	n = DUK_HCOMPILEDFUNCTION_GET_FUNCS_COUNT(heap, fun);
+	n = DUK_HCOMPFUNC_GET_FUNCS_COUNT(heap, fun);
 	duk_debug_write_int(thr, (duk_int32_t) n);
-	fn = DUK_HCOMPILEDFUNCTION_GET_FUNCS_BASE(heap, fun);
+	fn = DUK_HCOMPFUNC_GET_FUNCS_BASE(heap, fun);
 	for (i = 0; i < n; i++) {
 		duk_debug_write_hobject(thr, *fn);
 		fn++;
 	}
 	duk_debug_write_string(thr,
-	                       (const char *) DUK_HCOMPILEDFUNCTION_GET_CODE_BASE(heap, fun),
-	                       (duk_size_t) DUK_HCOMPILEDFUNCTION_GET_CODE_SIZE(heap, fun));
+	                       (const char *) DUK_HCOMPFUNC_GET_CODE_BASE(heap, fun),
+	                       (duk_size_t) DUK_HCOMPFUNC_GET_CODE_SIZE(heap, fun));
 	duk_debug_write_eom(thr);
 	return;
 
@@ -1886,8 +1888,8 @@ DUK_LOCAL void duk__debug_handle_get_bytecode(duk_hthread *thr, duk_heap *heap)
 	duk_debug_write_error_eom(thr, DUK_DBG_ERR_UNKNOWN, "invalid argument");
 	return;
 
- fail_level:
-	duk_debug_write_error_eom(thr, DUK_DBG_ERR_NOTFOUND, "invalid callstack level");
+ fail_index:
+	duk_debug_write_error_eom(thr, DUK_DBG_ERR_NOTFOUND, "invalid callstack index");
 	return;
 }
 
@@ -1919,7 +1921,8 @@ DUK_LOCAL duk_uint_t duk__debug_getinfo_heaphdr_masks[] = {
 DUK_LOCAL const char * const duk__debug_getinfo_hstring_keys[] = {
 #if 0
 	"arridx",
-	"internal",
+	"symbol",
+	"hidden",
 	"reserved_word",
 	"strict_reserved_word",
 	"eval_or_arguments",
@@ -1930,7 +1933,8 @@ DUK_LOCAL const char * const duk__debug_getinfo_hstring_keys[] = {
 DUK_LOCAL duk_uint_t duk__debug_getinfo_hstring_masks[] = {
 #if 0
 	DUK_HSTRING_FLAG_ARRIDX,
-	DUK_HSTRING_FLAG_INTERNAL,
+	DUK_HSTRING_FLAG_SYMBOL,
+	DUK_HSTRING_FLAG_HIDDEN,
 	DUK_HSTRING_FLAG_RESERVED_WORD,
 	DUK_HSTRING_FLAG_STRICT_RESERVED_WORD,
 	DUK_HSTRING_FLAG_EVAL_OR_ARGUMENTS,
@@ -1941,10 +1945,10 @@ DUK_LOCAL duk_uint_t duk__debug_getinfo_hstring_masks[] = {
 DUK_LOCAL const char * const duk__debug_getinfo_hobject_keys[] = {
 	"extensible",
 	"constructable",
-	"bound",
-	"compiledfunction",
-	"nativefunction",
-	"bufferobject",
+	"boundfunc",
+	"compfunc",
+	"natfunc",
+	"bufobj",
 	"thread",
 	"array_part",
 	"strict",
@@ -1963,10 +1967,10 @@ DUK_LOCAL const char * const duk__debug_getinfo_hobject_keys[] = {
 DUK_LOCAL duk_uint_t duk__debug_getinfo_hobject_masks[] = {
 	DUK_HOBJECT_FLAG_EXTENSIBLE,
 	DUK_HOBJECT_FLAG_CONSTRUCTABLE,
-	DUK_HOBJECT_FLAG_BOUND,
-	DUK_HOBJECT_FLAG_COMPILEDFUNCTION,
-	DUK_HOBJECT_FLAG_NATIVEFUNCTION,
-	DUK_HOBJECT_FLAG_BUFFEROBJECT,
+	DUK_HOBJECT_FLAG_BOUNDFUNC,
+	DUK_HOBJECT_FLAG_COMPFUNC,
+	DUK_HOBJECT_FLAG_NATFUNC,
+	DUK_HOBJECT_FLAG_BUFOBJ,
 	DUK_HOBJECT_FLAG_THREAD,
 	DUK_HOBJECT_FLAG_ARRAY_PART,
 	DUK_HOBJECT_FLAG_STRICT,
@@ -2070,8 +2074,11 @@ DUK_LOCAL duk_bool_t duk__debug_getprop_index(duk_hthread *thr, duk_heap *heap,
 	}
 
 	flags = DUK_HOBJECT_E_GET_FLAGS(heap, h_obj, idx);
-	if (DUK_HSTRING_HAS_INTERNAL(h_key)) {
-		flags |= DUK_DBG_PROPFLAG_INTERNAL;
+	if (DUK_HSTRING_HAS_SYMBOL(h_key)) {
+		flags |= DUK_DBG_PROPFLAG_SYMBOL;
+	}
+	if (DUK_HSTRING_HAS_HIDDEN(h_key)) {
+		flags |= DUK_DBG_PROPFLAG_HIDDEN;
 	}
 	duk_debug_write_uint(thr, flags);
 	duk_debug_write_hstring(thr, h_key);
@@ -2174,14 +2181,21 @@ DUK_LOCAL void duk__debug_handle_get_heap_obj_info(duk_hthread *thr, duk_heap *h
 		duk__debug_getinfo_prop_uint(thr, "a_size", (duk_uint_t) DUK_HOBJECT_GET_ASIZE(h_obj));
 		duk__debug_getinfo_prop_uint(thr, "h_size", (duk_uint_t) DUK_HOBJECT_GET_HSIZE(h_obj));
 
-		/* duk_hnativefunction specific fields. */
-		if (DUK_HOBJECT_IS_NATIVEFUNCTION(h_obj)) {
-			duk_hnativefunction *h_fun;
-			h_fun = (duk_hnativefunction *) h_obj;
+		if (DUK_HOBJECT_IS_ARRAY(h_obj)) {
+			duk_harray *h_arr;
+			h_arr = (duk_harray *) h_obj;
+
+			duk__debug_getinfo_prop_int(thr, "length", h_arr->length);
+			duk__debug_getinfo_prop_bool(thr, "length_nonwritable", h_arr->length_nonwritable);
+		}
+
+		if (DUK_HOBJECT_IS_NATFUNC(h_obj)) {
+			duk_hnatfunc *h_fun;
+			h_fun = (duk_hnatfunc *) h_obj;
 
 			duk__debug_getinfo_prop_int(thr, "nargs", h_fun->nargs);
 			duk__debug_getinfo_prop_int(thr, "magic", h_fun->magic);
-			duk__debug_getinfo_prop_bool(thr, "varargs", h_fun->magic == DUK_HNATIVEFUNCTION_NARGS_VARARGS);
+			duk__debug_getinfo_prop_bool(thr, "varargs", h_fun->magic == DUK_HNATFUNC_NARGS_VARARGS);
 			/* Native function pointer may be different from a void pointer,
 			 * and we serialize it from memory directly now (no byte swapping etc).
 			 */
@@ -2189,16 +2203,34 @@ DUK_LOCAL void duk__debug_handle_get_heap_obj_info(duk_hthread *thr, duk_heap *h
 			duk_debug_write_buffer(thr, (const char *) &h_fun->func, sizeof(h_fun->func));
 		}
 
-		if (DUK_HOBJECT_IS_COMPILEDFUNCTION(h_obj)) {
-			duk_hcompiledfunction *h_fun;
+		if (DUK_HOBJECT_IS_COMPFUNC(h_obj)) {
+			duk_hcompfunc *h_fun;
 			duk_hbuffer *h_buf;
-			h_fun = (duk_hcompiledfunction *) h_obj;
+			duk_hobject *h_lexenv;
+			duk_hobject *h_varenv;
+			h_fun = (duk_hcompfunc *) h_obj;
 
 			duk__debug_getinfo_prop_int(thr, "nregs", h_fun->nregs);
 			duk__debug_getinfo_prop_int(thr, "nargs", h_fun->nargs);
+
+			duk__debug_getinfo_flags_key(thr, "lex_env");
+			h_lexenv = DUK_HCOMPFUNC_GET_LEXENV(thr->heap, h_fun);
+			if (h_lexenv != NULL) {
+				duk_debug_write_hobject(thr, h_lexenv);
+			} else {
+				duk_debug_write_null(thr);
+			}
+			duk__debug_getinfo_flags_key(thr, "var_env");
+			h_varenv = DUK_HCOMPFUNC_GET_VARENV(thr->heap, h_fun);
+			if (h_varenv != NULL) {
+				duk_debug_write_hobject(thr, h_varenv);
+			} else {
+				duk_debug_write_null(thr);
+			}
+
 			duk__debug_getinfo_prop_uint(thr, "start_line", h_fun->start_line);
 			duk__debug_getinfo_prop_uint(thr, "end_line", h_fun->end_line);
-			h_buf = (duk_hbuffer *) DUK_HCOMPILEDFUNCTION_GET_DATA(thr->heap, h_fun);
+			h_buf = (duk_hbuffer *) DUK_HCOMPFUNC_GET_DATA(thr->heap, h_fun);
 			if (h_buf != NULL) {
 				duk__debug_getinfo_flags_key(thr, "data");
 				duk_debug_write_heapptr(thr, (duk_heaphdr *) h_buf);
@@ -2214,20 +2246,22 @@ DUK_LOCAL void duk__debug_handle_get_heap_obj_info(duk_hthread *thr, duk_heap *h
 			DUK_UNREF(h_thr);
 		}
 
-		if (DUK_HOBJECT_IS_BUFFEROBJECT(h_obj)) {
-			duk_hbufferobject *h_bufobj;
-			h_bufobj = (duk_hbufferobject *) h_obj;
+#if defined(DUK_USE_BUFFEROBJECT_SUPPORT)
+		if (DUK_HOBJECT_IS_BUFOBJ(h_obj)) {
+			duk_hbufobj *h_bufobj;
+			h_bufobj = (duk_hbufobj *) h_obj;
 
 			duk__debug_getinfo_prop_uint(thr, "slice_offset", h_bufobj->offset);
 			duk__debug_getinfo_prop_uint(thr, "slice_length", h_bufobj->length);
 			duk__debug_getinfo_prop_uint(thr, "elem_shift", (duk_uint_t) h_bufobj->shift);
 			duk__debug_getinfo_prop_uint(thr, "elem_type", (duk_uint_t) h_bufobj->elem_type);
-			duk__debug_getinfo_prop_bool(thr, "is_view", (duk_uint_t) h_bufobj->is_view);
+			duk__debug_getinfo_prop_bool(thr, "is_typedarray", (duk_uint_t) h_bufobj->is_typedarray);
 			if (h_bufobj->buf != NULL) {
 				duk__debug_getinfo_flags_key(thr, "buffer");
 				duk_debug_write_heapptr(thr, (duk_heaphdr *) h_bufobj->buf);
 			}
 		}
+#endif  /* DUK_USE_BUFFEROBJECT_SUPPORT */
 		break;
 	}
 	case DUK_HTYPE_BUFFER: {
@@ -2576,7 +2610,7 @@ DUK_INTERNAL duk_bool_t duk_debug_process_messages(duk_hthread *thr, duk_bool_t
 			break;
 		}
 
-		if (!thr->heap->dbg_paused || no_block) {
+		if (!DUK_HEAP_HAS_DEBUGGER_PAUSED(thr->heap) || no_block) {
 			if (!duk_debug_read_peek(thr)) {
 				/* Note: peek cannot currently trigger a detach
 				 * so the dbg_detaching == 0 assert outside the
@@ -2629,7 +2663,7 @@ DUK_INTERNAL duk_bool_t duk_debug_process_messages(duk_hthread *thr, duk_bool_t
 
 DUK_INTERNAL void duk_debug_halt_execution(duk_hthread *thr, duk_bool_t use_prev_pc) {
 	duk_activation *act;
-	duk_hcompiledfunction *fun;
+	duk_hcompfunc *fun;
 	duk_instr_t *old_pc = NULL;
 
 	DUK_ASSERT(thr != NULL);
@@ -2649,16 +2683,16 @@ DUK_INTERNAL void duk_debug_halt_execution(duk_hthread *thr, duk_bool_t use_prev
 	if (act != NULL) {
 		duk_hthread_sync_currpc(thr);
 		old_pc = act->curr_pc;
-		fun = (duk_hcompiledfunction *) DUK_ACT_GET_FUNC(act);
+		fun = (duk_hcompfunc *) DUK_ACT_GET_FUNC(act);
 
 		/* Short circuit if is safe: if act->curr_pc != NULL, 'fun' is
 		 * guaranteed to be a non-NULL Ecmascript function.
 		 */
 		DUK_ASSERT(act->curr_pc == NULL ||
-		           (fun != NULL && DUK_HOBJECT_IS_COMPILEDFUNCTION((duk_hobject *) fun)));
+		           (fun != NULL && DUK_HOBJECT_IS_COMPFUNC((duk_hobject *) fun)));
 		if (use_prev_pc &&
 		    act->curr_pc != NULL &&
-		    act->curr_pc > DUK_HCOMPILEDFUNCTION_GET_CODE_BASE(thr->heap, fun)) {
+		    act->curr_pc > DUK_HCOMPFUNC_GET_CODE_BASE(thr->heap, fun)) {
 			act->curr_pc--;
 		}
 	}
@@ -2671,7 +2705,7 @@ DUK_INTERNAL void duk_debug_halt_execution(duk_hthread *thr, duk_bool_t use_prev
 	 */
 
 	thr->heap->dbg_state_dirty = 1;
-	while (thr->heap->dbg_paused) {
+	while (DUK_HEAP_HAS_DEBUGGER_PAUSED(thr->heap)) {
 		DUK_ASSERT(DUK_HEAP_IS_DEBUGGER_ATTACHED(thr->heap));
 		DUK_ASSERT(thr->heap->dbg_processing);
 		duk_debug_process_messages(thr, 0 /*no_block*/);
@@ -2764,8 +2798,6 @@ DUK_INTERNAL duk_bool_t duk_debug_remove_breakpoint(duk_hthread *thr, duk_small_
 	return 1;
 }
 
-#undef DUK__SET_CONN_BROKEN
-
 #else  /* DUK_USE_DEBUGGER_SUPPORT */
 
 /* No debugger support. */
